# Unit test add_y_train_argument
# ==============================================================================
import inspect
import numpy as np
from skforecast.metrics import add_y_train_argument


def test_add_y_train_argument_func_no_y_train():
    """
    Test add_y_train_argument when the function does not have the 'y_train' argument.
    """
    def test_func(x):
        return x

    wrapped_func = add_y_train_argument(test_func)

    # Check if the wrapped function has the 'y_train' argument
    assert 'y_train' in inspect.signature(wrapped_func).parameters

    # Test the wrapped function with 'y_train' argument
    result = wrapped_func(10, y_train=[1, 2, 3])
    assert result == 10

    # Test the wrapped function without 'y_train' argument
    result = wrapped_func(10)
    assert result == 10


def test_add_y_train_argument_func_with_y_train():
    """ 
    Test add_y_train_argument when the function already has the 'y_train' argument.
    """
    def test_func(x, y_train):
        return x + y_train
    
    wrapped_func = add_y_train_argument(test_func)

    # Check if the wrapped function has the 'y_train' argument
    assert 'y_train' in inspect.signature(wrapped_func).parameters

    # Test the wrapped function with 'y_train' argument
    result = wrapped_func(10, y_train=np.array([1, 2, 3]))
    np.testing.assert_array_almost_equal(result, np.array([11, 12, 13]))


# Unit test _get_metric
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.metrics import (
    mean_squared_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_log_error,
    median_absolute_error,
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    precision_score,
    recall_score
)

from skforecast.metrics import _get_metric
from skforecast.metrics import (
    mean_absolute_scaled_error,
    root_mean_squared_scaled_error,
    symmetric_mean_absolute_percentage_error
)


def test_get_metric_ValueError_when_metric_not_in_metrics_allowed():
    """
    Test ValueError is raised when metric is not in metrics allowed.
    """
    metric = 'not_a_metric'

    allowed_metrics = [
        # Regression metrics
        "mean_squared_error",
        "mean_absolute_error",
        "mean_absolute_percentage_error",
        "mean_squared_log_error",
        "mean_absolute_scaled_error",
        "root_mean_squared_scaled_error",
        "median_absolute_error",
        "symmetric_mean_absolute_percentage_error",

        # Classification metrics
        "accuracy_score",
        "balanced_accuracy_score",
        "f1_score",
        "precision_score",
        "recall_score"
    ]
    
    err_msg = re.escape(f"Allowed metrics are: {allowed_metrics}. Got {metric}.")
    with pytest.raises(ValueError, match = err_msg):
        _get_metric(metric)


@pytest.mark.parametrize("metric_str, metric_callable", 
                         [('mean_squared_error', mean_squared_error),
                          ('mean_absolute_error', mean_absolute_error),
                          ('mean_absolute_percentage_error', mean_absolute_percentage_error),
                          ('mean_squared_log_error', mean_squared_log_error),
                          ('median_absolute_error', median_absolute_error),
                          ('symmetric_mean_absolute_percentage_error', symmetric_mean_absolute_percentage_error)], 
                         ids = lambda dt: f'mertic_str, metric_callable: {dt}')
def test_get_metric_output_for_all_metrics(metric_str, metric_callable):
    """
    Test output for all metrics allowed.
    """
    y_true = np.array([3, 1, 2, 7])
    y_pred = np.array([2.5, 0.0, 2, 8])

    metric = _get_metric(metric_str)
    expected = metric_callable(y_true=y_true, y_pred=y_pred)
    
    assert metric(y_true=y_true, y_pred=y_pred) == expected


@pytest.mark.parametrize("metric_str, metric_callable", 
                         [('accuracy_score', accuracy_score),
                          ('balanced_accuracy_score', balanced_accuracy_score),
                          ('f1_score', f1_score),
                          ('precision_score', precision_score),
                          ('recall_score', recall_score)], 
                         ids = lambda dt: f'mertic_str, metric_callable: {dt}')
def test_get_metric_output_for_all_metrics_classification(metric_str, metric_callable):
    """
    Test output for all metrics allowed.
    """
    y_true = np.array([0, 1, 1, 0, 1, 0])
    y_pred = np.array([0, 0, 1, 0, 0, 1])

    metric = _get_metric(metric_str)
    expected = metric_callable(y_true=y_true, y_pred=y_pred)
    
    assert metric(y_true=y_true, y_pred=y_pred) == expected


@pytest.mark.parametrize("metric_str, metric_callable", 
                         [('mean_absolute_scaled_error', mean_absolute_scaled_error),
                          ('root_mean_squared_scaled_error', root_mean_squared_scaled_error)], 
                         ids = lambda dt: f'mertic_str, metric_callable: {dt}')
def test_get_metric_output_for_all_metrics_y_train(metric_str, metric_callable):
    """
    Test output for all metrics allowed with y_train argument.
    """
    y_true = np.array([3, -0.5, 2, 7])
    y_pred = np.array([2.5, 0.0, 2, 8])
    y_train = np.array([2, 3, 4, 5])

    metric = _get_metric(metric_str)
    expected = metric_callable(y_true=y_true, y_pred=y_pred, y_train=y_train)
    
    assert metric(y_true=y_true, y_pred=y_pred, y_train=y_train) == expected



# Unit test _check_categorical_support ForecasterRecursiveClassifier
# ==============================================================================
import pytest
from lightgbm import LGBMClassifier
from sklearn.calibration import CalibratedClassifierCV
from sklearn.linear_model import LogisticRegression
from skforecast.recursive import ForecasterRecursiveClassifier


@pytest.mark.parametrize("estimator", 
                         [LogisticRegression(), 
                          CalibratedClassifierCV(LogisticRegression())], 
                         ids = lambda estimator: f'estimator: {estimator}')
def test_check_categorical_support_LogisticRegression(estimator):
    """
    Test _check_categorical_support method with LogisticRegression.
    """
    forecaster = ForecasterRecursiveClassifier(
        estimator         = estimator,
        lags              = 3,
        features_encoding = 'auto'
    )

    results = forecaster._check_categorical_support(estimator=forecaster.estimator)
    assert results is False


@pytest.mark.parametrize("estimator", 
                         [LGBMClassifier(), 
                          CalibratedClassifierCV(LGBMClassifier())], 
                         ids = lambda estimator: f'estimator: {estimator}')
def test_check_categorical_support_LGBMClassifier(estimator):
    """
    Test _check_categorical_support method with LGBMClassifier.
    """
    forecaster = ForecasterRecursiveClassifier(
        estimator         = estimator,
        lags              = 3,
        features_encoding = 'auto'
    )

    results = forecaster._check_categorical_support(estimator=forecaster.estimator)
    assert results is True



# Unit test plot_residuals
# ==============================================================================
import re
import pytest
from skforecast.plot import plot_residuals


def test_plot_residuals_ValueError_when_residuals_are_not_provided():
    """
    Test ValueError if `residuals` argument is None and also `y_true` and `y_pred`
    are None.
    """
    err_msg = re.escape(
        "If `residuals` argument is None then, `y_true` and `y_pred` must be provided."
    )
    with pytest.raises(ValueError, match=err_msg):
        plot_residuals()



from ._split import (
    TimeSeriesFold,
    OneStepAheadFold
)
from ._validation import (
    backtesting_forecaster,
    backtesting_forecaster_multiseries,
    backtesting_sarimax,
    backtesting_stats
)
from ._search import (
    grid_search_forecaster,
    random_search_forecaster,
    bayesian_search_forecaster,
    grid_search_forecaster_multiseries,
    random_search_forecaster_multiseries,
    bayesian_search_forecaster_multiseries,
    grid_search_sarimax,
    random_search_sarimax,
    grid_search_stats,
    random_search_stats
)


# Unit test predict ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import FunctionTransformer
from sklearn.compose import make_column_transformer
from sklearn.compose import make_column_selector
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor

from skforecast.preprocessing import RollingFeatures
from skforecast.preprocessing import TimeSeriesDifferentiator
from skforecast.direct import ForecasterDirectMultiVariate

# Fixtures
from .fixtures_forecaster_direct_multivariate import series
from .fixtures_forecaster_direct_multivariate import exog
from .fixtures_forecaster_direct_multivariate import exog_predict
from .fixtures_forecaster_direct_multivariate import data  # to test results when using differentiation
from ....recursive.tests.tests_forecaster_recursive_multiseries.fixtures_forecaster_recursive_multiseries import expected_df_to_long_format

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


@pytest.mark.parametrize("steps", [[1, 2.0, 3], [1, 4.]], 
                         ids=lambda steps: f'steps: {steps}')
def test_predict_TypeError_when_steps_list_contain_floats(steps):
    """
    Test predict TypeError when steps is a list with floats.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=3
    )
    forecaster.fit(series=series)

    err_msg = re.escape(
        f"`steps` argument must be an int, a list of ints or `None`. "
        f"Got {type(steps)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.predict(steps=steps)


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=3
    )

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict(steps=5)


@pytest.mark.parametrize("steps", [3, [1, 2, 3], None], 
                         ids=lambda steps: f'steps: {steps}')
def test_predict_output_when_estimator_is_LinearRegression(steps):
    """
    Test predict output when using LinearRegression as estimator.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=3
    )
    forecaster.fit(series=series)
    results = forecaster.predict(steps=steps)
    expected = pd.DataFrame(
                   data    = np.array([0.63114259, 0.3800417, 0.33255977]),
                   index   = pd.RangeIndex(start=50, stop=53, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_with_list_interspersed():
    """
    Test predict output when using LinearRegression as estimator and steps is
    a list with interspersed steps.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l2', lags=3, steps=5
    )
    forecaster.fit(series=series)
    results = forecaster.predict(steps=[1, 4])
    expected = pd.DataFrame(
                   data    = np.array([0.61048324, 0.53962565]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1)[[0, 3]],
                   columns = ['l2']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_with_different_lags():
    """
    Test predict output when using LinearRegression as estimator and different
    lags configuration for each series.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l2', lags={'l1': 5, 'l2': [1, 7]}, steps=3
    )
    forecaster.fit(series=series)
    results = forecaster.predict(steps=3)
    expected = pd.DataFrame(
                   data    = np.array([0.58053278, 0.43052971, 0.60582844]),
                   index   = pd.RangeIndex(start=50, stop=53, step=1),
                   columns = ['l2']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_with_lags_dict_with_None_in_level_lags():
    """
    Test predict output when using LinearRegression as estimator when lags is a 
    dict and level has None lags configuration.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l2', lags={'l1': 5, 'l2': None}, steps=3
    )
    forecaster.fit(series=series)
    results = forecaster.predict(steps=3)
    expected = pd.DataFrame(
                   data    = np.array([0.59695170, 0.49497765, 0.51970720]),
                   index   = pd.RangeIndex(start=50, stop=53, step=1),
                   columns = ['l2']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_with_lags_dict_with_None_but_no_in_level():
    """
    Test predict output when using LinearRegression as estimator when lags is a 
    dict with None values.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags={'l1': 5, 'l2': None}, steps=3
    )
    forecaster.fit(series=series)
    results = forecaster.predict(steps=3)
    expected = pd.DataFrame(
                   data    = np.array([0.61119488, 0.48858659, 0.46753222]),
                   index   = pd.RangeIndex(start=50, stop=53, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_using_last_window():
    """
    Test predict output when using LinearRegression as estimator and last_window.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=3
    )
    forecaster.fit(series=series)
    last_window = pd.DataFrame(
                      data    = np.array([[0.98555979, 0.39887629],
                                          [0.51948512, 0.2408559 ],
                                          [0.61289453, 0.34345601]]),
                      index   = pd.RangeIndex(start=47, stop=50, step=1),
                      columns = ['l1', 'l2']
                  )
    results = forecaster.predict(steps=[1, 2], last_window=last_window)
    expected = pd.DataFrame(
                   data    = np.array([0.63114259, 0.3800417]),
                   index   = pd.RangeIndex(start=50, stop=52, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_using_exog():
    """
    Test predict output when using LinearRegression as estimator and exog.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=3
    )
    forecaster.fit(series=series.iloc[:40,], exog=exog.iloc[:40, 0])
    results = forecaster.predict(steps=None, exog=exog.iloc[40:43, 0])
    expected = pd.DataFrame(
                   data    = np.array([0.57243323, 0.56121924, 0.38879785]),
                   index   = pd.RangeIndex(start=40, stop=43, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_with_transform_series():
    """
    Test predict output when using LinearRegression as estimator and StandardScaler.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     lags               = 5,
                     steps              = 5,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series)
    results = forecaster.predict()
    expected = pd.DataFrame(
                   data    = np.array([0.60056539, 0.42924504, 0.34173573, 0.44231236, 0.40133213]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_with_transform_series_as_dict():
    """
    Test predict output when using LinearRegression as estimator and transformer_series
    is a dict with 2 different transformers.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l2',
                     lags               = 5,
                     steps              = 5,
                     transformer_series = {'l1': StandardScaler(), 'l2': MinMaxScaler()}
                 )
    forecaster.fit(series=series)
    results = forecaster.predict()
    expected = pd.DataFrame(
                   data    = np.array([0.65049981, 0.57548048, 0.64278726, 0.54421867, 0.7851753]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['l2']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_predict_output_when_transform_series_and_transform_exog(n_jobs):
    """
    Test predict output when using LinearRegression as estimator, StandardScaler
    as transformer_series and transformer_exog as transformer_exog.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     lags               = 5,
                     steps              = 5,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                     n_jobs             = n_jobs
                 )
    forecaster.fit(series=series, exog=exog)
    results = forecaster.predict(steps=[1, 2, 3, 4, 5], exog=exog_predict)
    expected = pd.DataFrame(
                   data    = np.array([0.61043227, 0.46658137, 0.54994519, 0.52561227, 0.46596527]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_and_weight_func():
    """
    Test predict output when using LinearRegression as estimator and custom_weights.
    """
    def custom_weights(index):
        """
        Return 1 for all elements in index
        """
        weights = np.ones_like(index)

        return weights
    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=3, weight_func=custom_weights
    )
    forecaster.fit(series=series)
    results = forecaster.predict(steps=3)
    expected = pd.DataFrame(
                   data    = np.array([0.63114259, 0.3800417, 0.33255977]),
                   index   = pd.RangeIndex(start=50, stop=53, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_output_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test predict output when using HistGradientBoostingRegressor and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog['exog_1'],
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = HistGradientBoostingRegressor(
                                              categorical_features = categorical_features,
                                              random_state         = 123
                                          ),
                     level              = 'l1',
                     lags               = 5,
                     steps              = 10,
                     transformer_series = None,
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
                   data = np.array([0.50131059, 0.49276926, 0.47433929, 0.4668392 , 
                                    0.47754412, 0.47360906, 0.47749396, 0.48461923, 
                                    0.48686681, 0.50223394]),
                   index   = pd.RangeIndex(start=50, stop=60, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMRegressor():
    """
    Test predict output when using LGBMRegressor and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog['exog_1'],
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LGBMRegressor(random_state=123),
                     level              = 'l1',
                     lags               = 5,
                     steps              = 10,
                     transformer_series = None,
                     transformer_exog   = transformer_exog,
                     fit_kwargs         = {'categorical_feature': categorical_features}
                 )
    forecaster.fit(series=series, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
                   data = np.array([0.50131059, 0.49276926, 0.47433929, 0.46683919,
                                    0.47754412, 0.47360906, 0.47749395, 0.48461923, 
                                    0.48686681, 0.50223394]),
                   index   = pd.RangeIndex(start=50, stop=60, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMRegressor_auto():
    """
    Test predict output when using LGBMRegressor and categorical variables with 
    categorical_features='auto'.
    """
    df_exog = pd.DataFrame({'exog_1': exog['exog_1'],
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    pipeline_categorical = make_pipeline(
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               FunctionTransformer(
                                   func=lambda x: x.astype('category'),
                                   feature_names_out= 'one-to-one'
                               )
                           )
    transformer_exog = make_column_transformer(
                            (
                                pipeline_categorical,
                                make_column_selector(dtype_exclude=np.number)
                            ),
                            remainder="passthrough",
                            verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LGBMRegressor(random_state=123),
                     level              = 'l1',
                     lags               = 5,
                     steps              = 10,
                     transformer_series = None,
                     transformer_exog   = transformer_exog,
                     fit_kwargs         = {'categorical_feature': 'auto'}
                 )
    forecaster.fit(series=series, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
                   data = np.array([0.50131059, 0.49276926, 0.47433929, 0.46683919,
                                    0.47754412, 0.47360906, 0.47749395, 0.48461923, 
                                    0.48686681, 0.50223394]),
                   index   = pd.RangeIndex(start=50, stop=60, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_with_exog_and_differentiation_is_1_steps_1():
    """
    Test predict output when using LinearRegression as estimator and 
    differentiation=1 and steps=1.
    """

    arr = data.to_numpy(copy=True)
    series_2 = pd.DataFrame(
        {'l1': arr,
         'l2': arr * 1.6},
        index=data.index
    )

    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    series_diff = pd.DataFrame(
        {'l1': differentiator.fit_transform(arr),
         'l2': differentiator.fit_transform(arr * 1.6)},
        index=data.index
    ).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=15, transformer_series=None
    )
    forecaster_1.fit(series=series_diff.loc[:end_train], exog=exog_diff.loc[:end_train])
    predictions_diff = forecaster_1.predict(exog=exog_diff.loc[end_train:])
    # Revert the differentiation
    predictions_diff = pd.pivot_table(
        predictions_diff, index=predictions_diff.index, columns="level", values="pred"
    )
    last_value_train = series_2[['l1']].loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff]).cumsum()[1:]
    predictions_1 = expected_df_to_long_format(predictions_1)

    forecaster_2 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=15, transformer_series=None, differentiation=1
    )
    forecaster_2.fit(series=series_2.loc[:end_train], exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(exog=exog.loc[end_train:])

    pd.testing.assert_frame_equal(predictions_1.asfreq('MS'), predictions_2)


def test_predict_output_when_with_exog_and_differentiation_is_1_steps_10():
    """
    Test predict output when using LinearRegression as estimator and 
    differentiation=1 and steps=10.
    """

    arr = data.to_numpy(copy=True)
    series_dt = pd.DataFrame(
        {'l1': arr,
         'l2': arr * 1.6},
        index=data.index
    )

    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    series_diff = pd.DataFrame(
        {'l1': differentiator.fit_transform(arr),
         'l2': differentiator.fit_transform(arr * 1.6)},
        index=data.index
    ).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog_dt = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog_dt.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=10, lags=15, transformer_series=None
    )
    forecaster_1.fit(series=series_diff.loc[:end_train], exog=exog_diff.loc[:end_train])
    predictions_diff = forecaster_1.predict(exog=exog_diff.loc[end_train:])
    # Revert the differentiation
    predictions_diff = pd.pivot_table(
        predictions_diff, index=predictions_diff.index, columns="level", values="pred"
    )
    last_value_train = series_dt[['l1']].loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff]).cumsum()[1:]
    predictions_1 = expected_df_to_long_format(predictions_1)

    forecaster_2 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=10, lags=15, transformer_series=None, differentiation=1
    )
    forecaster_2.fit(series=series_dt.loc[:end_train], exog=exog_dt.loc[:end_train])
    predictions_2 = forecaster_2.predict(exog=exog_dt.loc[end_train:])

    pd.testing.assert_frame_equal(predictions_1.asfreq('MS'), predictions_2)


def test_predict_output_when_with_exog_and_differentiation_is_2():
    """
    Test predict output when using LinearRegression as estimator and differentiation=2.
    """

    arr = data.to_numpy(copy=True)
    series_dt = pd.DataFrame(
        {'l1': arr,
         'l2': arr * 1.6},
        index=data.index
    )

    # Data differentiated
    df_diff_1 = pd.DataFrame()
    df_diff_2 = pd.DataFrame()
    for col in series_dt.columns:

        differentiator_1 = TimeSeriesDifferentiator(order=1)
        differentiator_2 = TimeSeriesDifferentiator(order=2)

        data_diff_1 = differentiator_1.fit_transform(series_dt[col].to_numpy())
        data_diff_1 = pd.Series(data_diff_1).dropna()
        df_diff_1[col] = data_diff_1

        data_diff_2 = differentiator_2.fit_transform(series_dt[col].to_numpy())
        data_diff_2 = pd.Series(data_diff_2).dropna()
        df_diff_2[col] = data_diff_2

    df_diff_1.index = series_dt.index[1:]
    df_diff_2.index = series_dt.index[2:]

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff_2 = exog.iloc[2:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=15, transformer_series=None
    )
    forecaster_1.fit(series=df_diff_2.loc[:end_train], exog=exog_diff_2.loc[:end_train])
    predictions_diff_2 = forecaster_1.predict(exog=exog_diff_2.loc[end_train:])
    
    # Revert the differentiation
    predictions_diff_2 = pd.pivot_table(
        predictions_diff_2, index=predictions_diff_2.index, columns="level", values="pred"
    )
    last_value_train_diff = df_diff_1[['l1']].loc[:end_train].iloc[[-1]]
    predictions_diff_1 = pd.concat([last_value_train_diff, predictions_diff_2]).cumsum()[1:]
    last_value_train = series_dt[['l1']].loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff_1]).cumsum()[1:]
    predictions_1 = expected_df_to_long_format(predictions_1)

    forecaster_2 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=15, 
        transformer_series=None, differentiation=2
    )
    forecaster_2.fit(series=series_dt.loc[:end_train], exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(exog=exog.loc[end_train:])

    pd.testing.assert_frame_equal(predictions_1.asfreq('MS'), predictions_2, check_names=False)


def test_predict_output_when_with_exog_and_differentiation_is_2_steps_10():
    """
    Test predict output when using LinearRegression as estimator and 
    differentiation=2 and steps=10.
    """

    arr = data.to_numpy(copy=True)
    series_dt = pd.DataFrame(
        {'l1': arr,
         'l2': arr * 1.6},
        index=data.index
    )

    # Data differentiated
    df_diff_1 = pd.DataFrame()
    df_diff_2 = pd.DataFrame()
    for col in series_dt.columns:

        differentiator_1 = TimeSeriesDifferentiator(order=1)
        differentiator_2 = TimeSeriesDifferentiator(order=2)

        data_diff_1 = differentiator_1.fit_transform(series_dt[col].to_numpy())
        data_diff_1 = pd.Series(data_diff_1).dropna()
        df_diff_1[col] = data_diff_1

        data_diff_2 = differentiator_2.fit_transform(series_dt[col].to_numpy())
        data_diff_2 = pd.Series(data_diff_2).dropna()
        df_diff_2[col] = data_diff_2

    df_diff_1.index = series_dt.index[1:]
    df_diff_2.index = series_dt.index[2:]

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff_2 = exog.iloc[2:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=10, lags=15, transformer_series=None
    )
    forecaster_1.fit(series=df_diff_2.loc[:end_train], exog=exog_diff_2.loc[:end_train])
    predictions_diff_2 = forecaster_1.predict(exog=exog_diff_2.loc[end_train:])

    # Revert the differentiation
    predictions_diff_2 = pd.pivot_table(
        predictions_diff_2, index=predictions_diff_2.index, columns="level", values="pred"
    )
    last_value_train_diff = df_diff_1[['l1']].loc[:end_train].iloc[[-1]]
    predictions_diff_1 = pd.concat([last_value_train_diff, predictions_diff_2]).cumsum()[1:]
    last_value_train = series_dt[['l1']].loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff_1]).cumsum()[1:]
    predictions_1 = expected_df_to_long_format(predictions_1)

    forecaster_2 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=10, lags=15, 
        transformer_series=None, differentiation=2
    )
    forecaster_2.fit(series=series_dt.loc[:end_train], exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(exog=exog.loc[end_train:])

    pd.testing.assert_frame_equal(predictions_1.asfreq('MS'), predictions_2)


def test_predict_output_when_window_features_steps_1():
    """
    Test output of predict when estimator is LGBMRegressor and window features
    with steps=1.
    """

    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirectMultiVariate(
        estimator=LGBMRegressor(verbose=-1, random_state=123), level='l1', 
        steps=1, lags=5, window_features=rolling
    )
    forecaster.fit(series=series, exog=exog['exog_1'])
    predictions = forecaster.predict(exog=exog_predict['exog_1'])

    expected = pd.DataFrame(
                   data = np.array([0.6005080563629931]),
                   index   = pd.RangeIndex(start=50, stop=51, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_window_features_steps_10():
    """
    Test output of predict when estimator is LGBMRegressor and window features
    with steps=10.
    """

    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirectMultiVariate(
        estimator=LGBMRegressor(verbose=-1, random_state=123), level='l1', 
        steps=10, lags=15, window_features=rolling
    )
    forecaster.fit(series=series, exog=exog['exog_1'])
    predictions = forecaster.predict(exog=exog_predict['exog_1'])

    expected = pd.DataFrame(
                   data = np.array([[0.50449652],
                                    [0.48055894],
                                    [0.48574331],
                                    [0.49495002],
                                    [0.50782532],
                                    [0.49700331],
                                    [0.49118152],
                                    [0.49641721],
                                    [0.48853374],
                                    [0.4886057 ]]),
                   index   = pd.RangeIndex(start=50, stop=60, step=1),
                   columns = ['l1']
               )
    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)



# Unit test grid_search_forecaster
# ==============================================================================
import numpy as np
import pandas as pd
import pytest
from sklearn.metrics import mean_absolute_percentage_error
from skforecast.metrics import mean_absolute_scaled_error
from skforecast.metrics import root_mean_squared_scaled_error
from sklearn.linear_model import Ridge
from sklearn.preprocessing import StandardScaler
from skforecast.recursive import ForecasterRecursive
from skforecast.direct import ForecasterDirect
from skforecast.model_selection._search import grid_search_forecaster
from skforecast.model_selection._split import TimeSeriesFold
from skforecast.model_selection._split import OneStepAheadFold

from tqdm import tqdm
from functools import partialmethod

tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)  # hide progress bar

# Fixtures
from ..fixtures_model_selection import y

 
def test_output_grid_search_forecaster_ForecasterRecursive_with_mocked():
    """
    Test output of grid_search_forecaster in ForecasterRecursive with mocked
    (mocked done in Skforecast v0.4.3)
    """
    forecaster = ForecasterRecursive(
        estimator=Ridge(random_state=123),
        lags=2,  # Placeholder, the value will be overwritten
    )
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = {"alpha": [0.01, 0.1, 1]}
    idx = len(lags_grid) * len(param_grid["alpha"])
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )

    results = grid_search_forecaster(
        forecaster=forecaster,
        y=y,
        cv=cv,
        lags_grid=lags_grid,
        param_grid=param_grid,
        metric="mean_squared_error",
        return_best=False,
        verbose=False,
    )

    expected_results = pd.DataFrame(
        {
            "lags": [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
            "lags_label": [
                [1, 2],
                [1, 2],
                [1, 2],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
            ],
            "params": [
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
            ],
            "mean_squared_error": np.array(
                [0.06464646, 0.06502362, 0.06745534, 0.06779272, 0.06802481, 0.06948609]
            ),
            "alpha": np.array([0.01, 0.1, 1.0, 0.01, 0.1, 1.0]),
        },
        index=pd.RangeIndex(start=0, stop=idx, step=1),
    )

    pd.testing.assert_frame_equal(results, expected_results)


forecasters = [
    ForecasterRecursive(estimator=Ridge(random_state=678), lags=3),
    ForecasterDirect(estimator=Ridge(random_state=678), lags=3, steps=1),
    ForecasterRecursive(
        estimator=Ridge(random_state=678),
        lags=3,
        transformer_y=StandardScaler(),
        transformer_exog=StandardScaler(),
    ),
    ForecasterDirect(
        estimator=Ridge(random_state=678),
        lags=3,
        steps=1,
        transformer_y=StandardScaler(),
        transformer_exog=StandardScaler(),
    ),
]
@pytest.mark.parametrize("forecaster", forecasters)
def test_grid_search_forecaster_equivalence_backtesting_one_step_ahead(
    forecaster,
):
    """
    Test that output of grid_search_forecaster is equivalent with backtesting and
    one step ahead one step=1, fixed_train_size=True retran=False
    """

    metrics = [
        "mean_absolute_error",
        "mean_squared_error",
        mean_absolute_percentage_error,
        mean_absolute_scaled_error,
        root_mean_squared_scaled_error,
    ]
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = {"alpha": [0.01, 0.1, 1]}
    cv_backtesting = TimeSeriesFold(
            steps                 = 1,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    cv_one_step_ahead = OneStepAheadFold(
            initial_train_size    = len(y_train),
            return_all_indexes    = False,
        )

    results_backtesting = grid_search_forecaster(
        forecaster=forecaster,
        y=y,
        cv=cv_backtesting,
        lags_grid=lags_grid,
        param_grid=param_grid,
        metric=metrics,
        return_best=False,
        verbose=False,
    )

    results_one_step_ahead = grid_search_forecaster(
        forecaster=forecaster,
        y=y,
        cv=cv_one_step_ahead,
        lags_grid=lags_grid,
        param_grid=param_grid,
        metric=metrics,
        return_best=False,
        verbose=False,
    )

    pd.testing.assert_frame_equal(results_backtesting, results_one_step_ahead)



# Unit test TimeSeriesDifferentiator init
# ==============================================================================
import re
import pytest
import numpy as np
from skforecast.preprocessing import TimeSeriesDifferentiator

# Fixtures
y = np.array([1, 4, 8, 10, 13, 22, 40, 46], dtype=float)
y_diff_1 = np.array([np.nan, 3., 4., 2., 3., 9., 18., 6.])
y_diff_2 = np.array([np.nan, np.nan, 1., -2., 1., 6., 9., -12.])
y_diff_3 = np.array([np.nan, np.nan, np.nan, -3., 3., 5., 3., -21.])

next_window = np.array([55, 70, 71], dtype=float)
next_window_diff_1 = np.array([9., 15., 1.], dtype=float)
next_window_diff_2 = np.array([3., 6., -14.], dtype=float)
next_window_diff_3 = np.array([15., 3., -20.], dtype=float)
next_window_series = np.concatenate([next_window, next_window], axis=0).reshape(2, -1).T
next_window_diff_1_2d = np.array([[9., 9.], [15., 15.], [1., 1.]], dtype=float)
next_window_diff_2_2d = np.array([[3., 3.], [6., 6.], [-14., -14.]], dtype=float)
next_window_diff_3_2d = np.array([[15., 15.], [3., 3.], [-20., -20.]], dtype=float)


def test_TimeSeriesDifferentiator_validate_params():
    """
    TimeSeriesDifferentiator validate params.
    """

    err_msg = re.escape(
        f"Parameter `order` must be an integer greater than 0. "
        f"Found {type(1.5)}."
    ) 
    with pytest.raises(TypeError, match = err_msg):
        TimeSeriesDifferentiator(order = 1.5)

    err_msg = re.escape(
        "Parameter `order` must be an integer greater than 0. "
        "Found 0."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        TimeSeriesDifferentiator(order = 0)

    err_msg = re.escape(
        f"Parameter `window_size` must be an integer greater than 0. "
        f"Found {type(1.5)}."
    ) 
    with pytest.raises(TypeError, match = err_msg):
        TimeSeriesDifferentiator(order = 1, window_size=1.5)

    err_msg = re.escape(
        "Parameter `window_size` must be an integer greater than 0. "
        "Found 0."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        TimeSeriesDifferentiator(order = 1, window_size=0)


def test_TimeSeriesDifferentiator_fit_order_1():
    """
    Test that TimeSeriesDifferentiator fit method with order 1.
    """
    X = np.arange(10)
    tsd = TimeSeriesDifferentiator(order=1, window_size=3)

    assert tsd.fit(X) == tsd
    assert tsd.initial_values == [0.]
    assert tsd.pre_train_values == [2.]
    assert tsd.last_values == [9.]


def test_TimeSeriesDifferentiator_fit_order_2():
    """
    Test that TimeSeriesDifferentiator fit method with order 2.
    """
    tsd = TimeSeriesDifferentiator(order=2, window_size=3)
    tsd.fit(y)

    assert tsd.fit(y) == tsd
    assert tsd.initial_values == [1., 3.]
    assert tsd.pre_train_values == [4., 4.]
    assert tsd.last_values == [46., 6.]


@pytest.mark.parametrize("order, expected",
                         [(1, y_diff_1),
                          (2, y_diff_2),
                          (3, y_diff_3)],
                         ids = lambda values: f'order, expected: {values}')
def test_TimeSeriesDifferentiator_transform(order, expected):
    """
    Test TimeSeriesDifferentiator transform method.
    """
    transformer = TimeSeriesDifferentiator(order=order)
    _ = transformer.fit(y)
    results = transformer.transform(y)

    np.testing.assert_array_almost_equal(results, expected)


@pytest.mark.parametrize("order, expected",
                         [(1, y_diff_1),
                          (2, y_diff_2),
                          (3, y_diff_3)],
                         ids = lambda values: f'order, expected: {values}')
def test_TimeSeriesDifferentiator_fit_transform(order, expected):
    """
    Test TimeSeriesDifferentiator fit_transform method.
    """
    transformer = TimeSeriesDifferentiator(order=order)
    results = transformer.fit_transform(y)

    np.testing.assert_array_almost_equal(results, expected)


@pytest.mark.parametrize("order",
                         [1, 2, 3],
                         ids = lambda values: f'order: {values}')
def test_TimeSeriesDifferentiator_inverse_transform(order):
    """
    Test TimeSeriesDifferentiator.inverse_transform method.
    """
    transformer = TimeSeriesDifferentiator(order=order)
    y_diff = transformer.fit_transform(y)
    results = transformer.inverse_transform(y_diff)

    np.testing.assert_array_almost_equal(results, y)


def test_TimeSeriesDifferentiator_inverse_transform_ValueError_no_window_size():
    """
    Test TimeSeriesDifferentiator inverse_transform_training ValueError when
    window_size is not provided.
    """
    transformer = TimeSeriesDifferentiator(order = 1)

    err_msg = re.escape(
        "The `window_size` parameter must be set before fitting the "
        "transformer to revert the differentiation of the training "
        "time series."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        transformer.inverse_transform_training(y_diff_1)        


@pytest.mark.parametrize("order, y_diff",
                         [(1, y_diff_1),
                          (2, y_diff_2),
                          (3, y_diff_3)],
                         ids = lambda values: f'order, y_diff: {values}')
def test_TimeSeriesDifferentiator_inverse_transform_training(order, y_diff):
    """
    Test TimeSeriesDifferentiator inverse_transform_training method.
    """
    window_size = 2 + order  # 2 lags + differentiation
    y_diff_train = y_diff[window_size:]  # Same as y_train predicted

    transformer = TimeSeriesDifferentiator(order=order, window_size=window_size)
    transformer.fit_transform(y)
    results = transformer.inverse_transform_training(y_diff_train)
    
    y_train_expected = y[window_size:]  # No differentiated

    assert len(y_train_expected) == len(y_diff_train)
    np.testing.assert_array_almost_equal(results, y_train_expected)


@pytest.mark.parametrize("order, next_window_diff, expected",
                         [(1, next_window_diff_1, next_window),
                          (2, next_window_diff_2, next_window),
                          (3, next_window_diff_3, next_window)],
                         ids = lambda values: f'order, next_window_diff, expected: {values}')
def test_TimeSeriesDifferentiator_inverse_transform_next_window(order, next_window_diff, expected):
    """
    Test TimeSeriesDifferentiator.inverse_transform_next_window method.
    """
    transformer = TimeSeriesDifferentiator(order=order)
    transformer.fit_transform(y)
    results = transformer.inverse_transform_next_window(next_window_diff)
    
    np.testing.assert_array_almost_equal(results, expected)


@pytest.mark.parametrize("order, next_window_diff, expected",
                         [(1, next_window_diff_1_2d, next_window_series),
                          (2, next_window_diff_2_2d, next_window_series),
                          (3, next_window_diff_3_2d, next_window_series)],
                         ids = lambda values: f'order, next_window_diff, expected: {values}')
def test_TimeSeriesDifferentiator_inverse_transform_next_window_2d(order, next_window_diff, expected):
    """
    Test TimeSeriesDifferentiator.inverse_transform_next_window method with 2d series.
    """
    transformer = TimeSeriesDifferentiator(order=order)
    transformer.fit_transform(y)
    results = transformer.inverse_transform_next_window(next_window_diff)
    
    np.testing.assert_array_almost_equal(results, expected)



# Unit test _bayesian_search_optuna
# ==============================================================================
import re
import pytest
import os
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import Ridge
from sklearn.metrics import mean_absolute_error, mean_absolute_percentage_error
from skforecast.exceptions import OneStepAheadValidationWarning
from skforecast.metrics import mean_absolute_scaled_error, root_mean_squared_scaled_error
from skforecast.recursive import ForecasterRecursive
from skforecast.direct import ForecasterDirect
from skforecast.model_selection import backtesting_forecaster
from skforecast.model_selection._search import _bayesian_search_optuna
from skforecast.model_selection._split import TimeSeriesFold, OneStepAheadFold
from skforecast.preprocessing import RollingFeatures
import optuna
from optuna.samplers import TPESampler
from tqdm import tqdm
from functools import partialmethod

# Fixtures
from ..fixtures_model_selection import y
from ..fixtures_model_selection import y_feature_selection
from ..fixtures_model_selection import exog_feature_selection

optuna.logging.set_verbosity(optuna.logging.WARNING)
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)  # hide progress bar


def test_TypeError_bayesian_search_optuna_when_cv_not_valid():
    """
    Test TypeError is raised in _bayesian_search_optuna when cv is not
    a valid splitter.
    """
    class DummyCV:
        pass

    cv = DummyCV()
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    
    def search_space(trial):  # pragma: no cover
        search_space  = {
            'alpha': trial.suggest_float('not_alpha', 1e-2, 1.0),
            'lags': trial.suggest_categorical('lags', [2, 4])
        }

        return search_space
    
    err_msg = re.escape(
        f"`cv` must be an instance of `TimeSeriesFold` or `OneStepAheadFold`. "
        f"Got {type(cv)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        _bayesian_search_optuna(
            forecaster         = forecaster,
            y                  = y,
            cv                 = cv,
            search_space       = search_space,
            metric             = ['mean_absolute_error', mean_absolute_error],
            n_trials           = 10,
            random_state       = 123,
            return_best        = False,
            verbose            = False,
        )


def test_ValueError_bayesian_search_optuna_metric_list_duplicate_names():
    """
    Test ValueError is raised in _bayesian_search_optuna when a `list` of 
    metrics is used with duplicate names.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):  # pragma: no cover
        search_space  = {
            'alpha': trial.suggest_float('not_alpha', 1e-2, 1.0),
            'lags': trial.suggest_categorical('lags', [2, 4])
        }

        return search_space

    err_msg = re.escape("When `metric` is a `list`, each metric name must be unique.")
    with pytest.raises(ValueError, match = err_msg):
        _bayesian_search_optuna(
            forecaster         = forecaster,
            y                  = y,
            cv                 = cv,
            search_space       = search_space,
            metric             = ['mean_absolute_error', mean_absolute_error],
            n_trials           = 10,
            random_state       = 123,
            return_best        = False,
            verbose            = False,
        )


def test_ValueError_bayesian_search_optuna_when_search_space_names_do_not_match():
    """
    Test ValueError is raised when search_space key name do not match the trial 
    object name from optuna.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2 
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    def search_space(trial):
        search_space  = {
            'alpha': trial.suggest_float('not_alpha', 1e-2, 1.0),
            'lags': trial.suggest_categorical('lags', [2, 4])
        }

        return search_space
    
    err_msg = re.escape(
        ("Some of the key values do not match the search_space key names.\n"
         "  Search Space keys  : ['alpha', 'lags']\n"
         "  Trial objects keys : ['not_alpha', 'lags'].")
    )
    with pytest.raises(ValueError, match = err_msg):
        _bayesian_search_optuna(
            forecaster         = forecaster,
            y                  = y,
            cv                 = cv, 
            search_space       = search_space,
            metric             = 'mean_absolute_error',
            n_trials           = 10,
            random_state       = 123,
            return_best        = False,
            verbose            = False,
        )


# This mark allows to only run test with "slow" label or all except this, "not slow".
# The mark should be included in the pytest.ini file
# pytest -m slow --verbose
# pytest -m "not slow" --verbose
@pytest.mark.slow
def test_results_output_bayesian_search_optuna_ForecasterRecursive():
    """
    Test output of _bayesian_search_optuna in ForecasterRecursive with mocked
    (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterRecursive(
                     estimator = RandomForestRegressor(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    def search_space(trial):
        search_space  = {
            'n_estimators'    : trial.suggest_int('n_estimators', 10, 20),
            'min_samples_leaf': trial.suggest_float('min_samples_leaf', 0.1, 1., log=True),
            'max_features'    : trial.suggest_categorical('max_features', ['log2', 'sqrt']),
            'lags'            : trial.suggest_categorical('lags', [2, 4])
        } 
        
        return search_space

    results = _bayesian_search_optuna(
                  forecaster         = forecaster,
                  y                  = y,
                  cv                 = cv,
                  search_space       = search_space,
                  metric             = 'mean_absolute_error',
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[np.array([1, 2, 3, 4]),
                {'n_estimators': 20, 'min_samples_leaf': 0.4839825891759374, 'max_features': 'log2'},
                0.21252324019730395, 20, 0.4839825891759374, 'log2'],
            [np.array([1, 2]),
                {'n_estimators': 15, 'min_samples_leaf': 0.34027307604369605, 'max_features': 'sqrt'},
                0.21479600790277778, 15, 0.34027307604369605, 'sqrt'],
            [np.array([1, 2]),
                {'n_estimators': 15, 'min_samples_leaf': 0.41010449151752726, 'max_features': 'sqrt'},
                0.21479600790277778, 15, 0.41010449151752726, 'sqrt'],
            [np.array([1, 2]),
                {'n_estimators': 14, 'min_samples_leaf': 0.782328520465639, 'max_features': 'log2'},
                0.21661388810185186, 14, 0.782328520465639, 'log2'],
            [np.array([1, 2]),
                {'n_estimators': 13, 'min_samples_leaf': 0.20142843988664705, 'max_features': 'sqrt'},
                0.22084665733716438, 13, 0.20142843988664705, 'sqrt'],
            [np.array([1, 2, 3, 4]),
                {'n_estimators': 17, 'min_samples_leaf': 0.21035794225904136, 'max_features': 'log2'},
                0.2229839747692736, 17, 0.21035794225904136, 'log2'],
            [np.array([1, 2]),
                {'n_estimators': 17, 'min_samples_leaf': 0.19325882509735576, 'max_features': 'sqrt'},
                0.2235827591941962, 17, 0.19325882509735576, 'sqrt'],
            [np.array([1, 2]),
                {'n_estimators': 11, 'min_samples_leaf': 0.2714570796881701, 'max_features': 'sqrt'},
                0.22384439522399655, 11, 0.2714570796881701, 'sqrt'],
            [np.array([1, 2, 3, 4]),
                {'n_estimators': 13, 'min_samples_leaf': 0.2599119286878713, 'max_features': 'log2'},
                0.22522461935181756, 13, 0.2599119286878713, 'log2'],
            [np.array([1, 2]),
                {'n_estimators': 14, 'min_samples_leaf': 0.1147302385573586, 'max_features': 'sqrt'},
                0.22764885273610677, 14, 0.1147302385573586, 'sqrt']],
            dtype=object),
        columns=['lags', 'params', 'mean_absolute_error', 'n_estimators', 'min_samples_leaf', 'max_features'],
        index=pd.Index([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype='int64')
    )

    pd.testing.assert_frame_equal(results, expected_results, check_dtype=False)


def test_results_output_bayesian_search_optuna_window_features_ForecasterRecursive():
    """
    Test output of _bayesian_search_optuna in ForecasterRecursive including
    window_features with mocked (mocked done in Skforecast v0.4.3).
    """
    window_features = RollingFeatures(
        stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation'],
        window_sizes = 3,
    )
    forecaster = ForecasterRecursive(
                     estimator = RandomForestRegressor(random_state=123),
                     lags      = 2,
                     window_features = window_features,
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    def search_space(trial):
        search_space  = {
            'n_estimators'    : trial.suggest_int('n_estimators', 10, 20),
            'min_samples_leaf': trial.suggest_float('min_samples_leaf', 0.1, 1., log=True),
            'max_features'    : trial.suggest_categorical('max_features', ['log2', 'sqrt']),
            'lags'            : trial.suggest_categorical('lags', [2, 4])
        } 
        
        return search_space

    results = _bayesian_search_optuna(
                  forecaster         = forecaster,
                  y                  = y,
                  cv                 = cv,
                  search_space       = search_space,
                  metric             = 'mean_absolute_error',
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False
              )[0]
    
    expected_results = pd.DataFrame({
        'lags': [
            np.array([1, 2, 3, 4]),
            np.array([1, 2]),
            np.array([1, 2]),
            np.array([1, 2]),
            np.array([1, 2, 3, 4]),
            np.array([1, 2, 3, 4]),
            np.array([1, 2]),
            np.array([1, 2]),
            np.array([1, 2]),
            np.array([1, 2]),
        ],
        'params': [
            {
                'n_estimators': 20,
                'min_samples_leaf': 0.4839825891759374,
                'max_features': 'log2',
            },
            {
                'n_estimators': 15,
                'min_samples_leaf': 0.41010449151752726,
                'max_features': 'sqrt',
            },
            {
                'n_estimators': 14,
                'min_samples_leaf': 0.782328520465639,
                'max_features': 'log2',
            },
            {
                'n_estimators': 15,
                'min_samples_leaf': 0.34027307604369605,
                'max_features': 'sqrt',
            },
            {
                'n_estimators': 13,
                'min_samples_leaf': 0.2599119286878713,
                'max_features': 'log2',
            },
            {
                'n_estimators': 17,
                'min_samples_leaf': 0.21035794225904136,
                'max_features': 'log2',
            },
            {
                'n_estimators': 11,
                'min_samples_leaf': 0.2714570796881701,
                'max_features': 'sqrt',
            },
            {
                'n_estimators': 13,
                'min_samples_leaf': 0.20142843988664705,
                'max_features': 'sqrt',
            },
            {
                'n_estimators': 17,
                'min_samples_leaf': 0.19325882509735576,
                'max_features': 'sqrt',
            },
            {
                'n_estimators': 14,
                'min_samples_leaf': 0.1147302385573586,
                'max_features': 'sqrt',
            },
        ],
        'mean_absolute_error': [
            0.21252324019730395,
            0.21851156682698414,
            0.22027505037414963,
            0.22530355563144466,
            0.2378358005415178,
            0.2384269598653063,
            0.25241649504238356,
            0.2584381014601676,
            0.26067982740527423,
            0.2697438465540914,
        ],
        'n_estimators': [20, 15, 14, 15, 13, 17, 11, 13, 17, 14],
        'min_samples_leaf': [
            0.4839825891759374,
            0.41010449151752726,
            0.782328520465639,
            0.34027307604369605,
            0.2599119286878713,
            0.21035794225904136,
            0.2714570796881701,
            0.20142843988664705,
            0.19325882509735576,
            0.1147302385573586,
        ],
        'max_features': [
            'log2',
            'sqrt',
            'log2',
            'sqrt',
            'log2',
            'log2',
            'sqrt',
            'sqrt',
            'sqrt',
            'sqrt',
        ],
    })

    pd.testing.assert_frame_equal(results, expected_results, check_dtype=False)
    

def test_results_output_bayesian_search_optuna_ForecasterRecursive_with_kwargs_create_study():
    """
    Test output of _bayesian_search_optuna in ForecasterRecursive with 
    kwargs_create_study with mocked (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    def search_space(trial):
        search_space  = {
            'alpha': trial.suggest_float('alpha', 1e-2, 1.0),
            'lags': trial.suggest_categorical('lags', [4, 2])
        }
        
        return search_space

    kwargs_create_study = {
        'sampler': TPESampler(seed=123, prior_weight=2.0, consider_magic_clip=False)
    }
    results = _bayesian_search_optuna(
                  forecaster          = forecaster,
                  y                   = y,
                  cv                  = cv, 
                  search_space        = search_space,
                  metric              = 'mean_absolute_error',
                  n_trials            = 10,
                  random_state        = 123,
                  return_best         = False,
                  verbose             = False,
                  kwargs_create_study = kwargs_create_study
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[np.array([1, 2]), {'alpha': 0.23598059857016607},
            0.21239141697571848, 0.23598059857016607],
        [np.array([1, 2]), {'alpha': 0.398196343012209}, 0.21271021033387602,
            0.398196343012209],
        [np.array([1, 2]), {'alpha': 0.4441865222328282}, 0.2127897499229874,
            0.4441865222328282],
        [np.array([1, 2]), {'alpha': 0.53623586010342}, 0.21293692257888705,
            0.53623586010342],
        [np.array([1, 2]), {'alpha': 0.7252189487445193},
            0.21319693043832985, 0.7252189487445193],
        [np.array([1, 2, 3, 4]), {'alpha': 0.9809565564007693},
            0.21539791166603497, 0.9809565564007693],
        [np.array([1, 2, 3, 4]), {'alpha': 0.8509374761370117},
            0.21557690844753197, 0.8509374761370117],
        [np.array([1, 2, 3, 4]), {'alpha': 0.7406154516747153},
            0.2157346392837304, 0.7406154516747153],
        [np.array([1, 2, 3, 4]), {'alpha': 0.6995044937418831},
            0.21579460210585208, 0.6995044937418831],
        [np.array([1, 2, 3, 4]), {'alpha': 0.5558016213920624},
            0.21600778429729228, 0.5558016213920624]], dtype=object),
        columns=['lags', 'params', 'mean_absolute_error', 'alpha'],
        index=pd.Index([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype='int64')
    ).astype({
        'mean_absolute_error': float,
        'alpha': float
    })

    pd.testing.assert_frame_equal(results, expected_results)


def test_results_output_bayesian_search_optuna_ForecasterRecursive_with_kwargs_study_optimize():
    """
    Test output of _bayesian_search_optuna in ForecasterRecursive when 
    kwargs_study_optimize with mocked (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterRecursive(
                     estimator = RandomForestRegressor(random_state=123),
                     lags      = 2 
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    def search_space(trial):
        search_space  = {
            'n_estimators': trial.suggest_int('n_estimators', 2, 10),
            'max_depth'   : trial.suggest_int('max_depth', 2, 10, log=True),
            'max_features': trial.suggest_categorical('max_features', ['log2', 'sqrt']),
            'lags'        : trial.suggest_categorical('lags', [2, 4])
        } 
        
        return search_space

    kwargs_study_optimize = {'timeout': 10}
    results = _bayesian_search_optuna(
                    forecaster            = forecaster,
                    y                     = y,
                    cv                    = cv,
                    search_space          = search_space,
                    metric                = 'mean_absolute_error',
                    n_trials              = 5,
                    random_state          = 123,
                    n_jobs                = 1,
                    return_best           = False,
                    verbose               = False,
                    kwargs_study_optimize = kwargs_study_optimize
                )[0].reset_index(drop=True)

    expected_results = pd.DataFrame(
        np.array([[np.array([1, 2, 3, 4]),
            {'n_estimators': 8, 'max_depth': 3, 'max_features': 'log2'},
            0.2176619102322017, 8, 3, 'log2'],
        [np.array([1, 2]),
            {'n_estimators': 8, 'max_depth': 3, 'max_features': 'sqrt'},
            0.21923614756760298, 8, 3, 'sqrt'],
        [np.array([1, 2]),
            {'n_estimators': 5, 'max_depth': 2, 'max_features': 'sqrt'},
            0.22116013675443522, 5, 2, 'sqrt'],
        [np.array([1, 2, 3, 4]),
            {'n_estimators': 10, 'max_depth': 6, 'max_features': 'log2'},
            0.22221487679563792, 10, 6, 'log2'],
        [np.array([1, 2]),
            {'n_estimators': 6, 'max_depth': 4, 'max_features': 'sqrt'},
            0.22883925084220677, 6, 4, 'sqrt']], dtype=object),
        columns=['lags', 'params', 'mean_absolute_error', 'n_estimators', 'max_depth', 'max_features'],
        index=pd.RangeIndex(start=0, stop=5, step=1)
    ).astype({
        'mean_absolute_error': float,
        'n_estimators': int,
        'max_depth': int
    })

    pd.testing.assert_frame_equal(results, expected_results, check_dtype=False)


def test_results_output_bayesian_search_optuna_ForecasterRecursive_when_lags_not_in_search_space():
    """
    Test output of _bayesian_search_optuna in ForecasterRecursive when lag is not 
    in search_space with mocked (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 4
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):
        search_space  = {'alpha': trial.suggest_float('alpha', 1e-2, 1.0)}

        return search_space

    results = _bayesian_search_optuna(
                  forecaster         = forecaster,
                  y                  = y,
                  search_space       = search_space,
                  cv                 = cv,
                  metric             = 'mean_absolute_error',
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[np.array([1, 2, 3, 4]), {'alpha': 0.9809565564007693},
            0.21539791166603497, 0.9809565564007693],
        [np.array([1, 2, 3, 4]), {'alpha': 0.7222742800877074},
            0.21576131952657338, 0.7222742800877074],
        [np.array([1, 2, 3, 4]), {'alpha': 0.6995044937418831},
            0.21579460210585208, 0.6995044937418831],
        [np.array([1, 2, 3, 4]), {'alpha': 0.6879814411990146},
            0.21581150916013203, 0.6879814411990146],
        [np.array([1, 2, 3, 4]), {'alpha': 0.5558016213920624},
            0.21600778429729228, 0.5558016213920624],
        [np.array([1, 2, 3, 4]), {'alpha': 0.48612258246951734},
            0.21611205459571634, 0.48612258246951734],
        [np.array([1, 2, 3, 4]), {'alpha': 0.42887539552321635},
            0.2161973389956996, 0.42887539552321635],
        [np.array([1, 2, 3, 4]), {'alpha': 0.398196343012209},
            0.21624265320052985, 0.398196343012209],
        [np.array([1, 2, 3, 4]), {'alpha': 0.29327794160087567},
            0.2163933942116072, 0.29327794160087567],
        [np.array([1, 2, 3, 4]), {'alpha': 0.2345829390285611},
            0.21647289061896782, 0.2345829390285611]], dtype=object),
        columns=['lags', 'params', 'mean_absolute_error', 'alpha'],
        index=pd.Index([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype='int64')
    ).astype({
        'mean_absolute_error': float,
        'alpha': float
    })
    
    pd.testing.assert_frame_equal(results, expected_results)


def test_evaluate_bayesian_search_optuna_when_return_best_ForecasterRecursive():
    """
    Test forecaster is refitted when return_best=True in _bayesian_search_optuna
    with a ForecasterRecursive.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):
        search_space  = {
            'alpha': trial.suggest_float('alpha', 1e-2, 1.0),
            'lags': trial.suggest_categorical('lags', [2, 4])
        }
        
        return search_space

    _bayesian_search_optuna(
        forecaster         = forecaster,
        y                  = y,
        cv                 = cv,
        search_space       = search_space,
        metric             = 'mean_absolute_error',
        n_trials           = 10,
        return_best        = True,
        verbose            = False
    )
    
    expected_lags = np.array([1, 2])
    expected_alpha = 0.5558016213920624

    np.testing.assert_array_almost_equal(forecaster.lags, expected_lags)
    assert expected_alpha == forecaster.estimator.alpha


def test_results_opt_best_output_bayesian_search_optuna_with_output_study_best_trial_optuna():
    """
    Test results_opt_best output of _bayesian_search_optuna with output 
    study.best_trial optuna.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )

    n_validation = 12
    y_train = y[:-n_validation]
    metric = 'mean_absolute_error'
    verbose = False
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    n_trials = 10
    random_state = 123

    def objective(
        trial,
        forecaster         = forecaster,
        y                  = y,
        cv                 = cv,
        metric             = metric,
        verbose            = verbose,
    ) -> float:
        
        alpha = trial.suggest_float('alpha', 1e-2, 1.0)
        lags  = trial.suggest_categorical('lags', [4, 2])
        
        forecaster = ForecasterRecursive(
                        estimator = Ridge(random_state=random_state, 
                                          alpha=alpha),
                        lags      = lags
                     )

        metric, _ = backtesting_forecaster(
                        forecaster         = forecaster,
                        y                  = y,
                        cv                 = cv,
                        metric             = metric,
                        verbose            = verbose       
                    )
        metric = metric.iat[0, 0]
        return metric
  
    study = optuna.create_study(direction="minimize", 
                                sampler=TPESampler(seed=random_state))
    study.optimize(objective, n_trials=n_trials)

    best_trial = study.best_trial

    def search_space(trial):
        search_space  = {
            'alpha': trial.suggest_float('alpha', 1e-2, 1.0),
            'lags': trial.suggest_categorical('lags', [4, 2])
        }
        return search_space
    return_best  = False

    results_opt_best = _bayesian_search_optuna(
                           forecaster         = forecaster,
                           y                  = y,
                           cv                 = cv,
                           search_space       = search_space,
                           metric             = metric,
                           n_trials           = n_trials,
                           return_best        = return_best,
                           verbose            = verbose
                       )[1]

    assert best_trial.number == results_opt_best.number
    assert best_trial.values == results_opt_best.values
    assert best_trial.params == results_opt_best.params


def test_results_output_bayesian_search_optuna_ForecasterDirect():
    """
    Test output of _bayesian_search_optuna in ForecasterDirect with mocked
    (mocked done in Skforecast v0.4.3).
    """    
    forecaster = ForecasterDirect(
                     estimator = RandomForestRegressor(random_state=123),
                     steps     = 3,
                     lags      = 4
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):
        search_space  = {
            'n_estimators'    : trial.suggest_int('n_estimators', 10, 20),
            'min_samples_leaf': trial.suggest_float('min_samples_leaf', 0.1, 1., log=True),
            'max_features'    : trial.suggest_categorical('max_features', ['log2', 'sqrt']),
            'lags'            : trial.suggest_categorical('lags', [2, 4])
        } 
        
        return search_space

    results = _bayesian_search_optuna(
                  forecaster   = forecaster,
                  y            = y,
                  cv           = cv,
                  search_space = search_space,
                  metric       = 'mean_absolute_error',
                  n_trials     = 10,
                  random_state = 123,
                  return_best  = False,
                  verbose      = False
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[np.array([1, 2, 3, 4]),
            {'n_estimators': 17, 'min_samples_leaf': 0.21035794225904136, 'max_features': 'log2'},
            0.20462849703549102, 17, 0.21035794225904136, 'log2'],
        [np.array([1, 2, 3, 4]),
            {'n_estimators': 13, 'min_samples_leaf': 0.2599119286878713, 'max_features': 'log2'},
            0.20911589421562574, 13, 0.2599119286878713, 'log2'],
        [np.array([1, 2, 3, 4]),
            {'n_estimators': 20, 'min_samples_leaf': 0.4839825891759374, 'max_features': 'log2'},
            0.2130714159765625, 20, 0.4839825891759374, 'log2'],
        [np.array([1, 2]),
            {'n_estimators': 15, 'min_samples_leaf': 0.34027307604369605, 'max_features': 'sqrt'},
            0.2133466144141166, 15, 0.34027307604369605, 'sqrt'],
        [np.array([1, 2]),
            {'n_estimators': 15, 'min_samples_leaf': 0.41010449151752726, 'max_features': 'sqrt'},
            0.21383764123529414, 15, 0.41010449151752726, 'sqrt'],
        [np.array([1, 2]),
            {'n_estimators': 14, 'min_samples_leaf': 0.782328520465639, 'max_features': 'log2'},
            0.21488066970063024, 14, 0.782328520465639, 'log2'],
        [np.array([1, 2]),
            {'n_estimators': 11, 'min_samples_leaf': 0.2714570796881701, 'max_features': 'sqrt'},
            0.21935169972870014, 11, 0.2714570796881701, 'sqrt'],
        [np.array([1, 2]),
            {'n_estimators': 13, 'min_samples_leaf': 0.20142843988664705, 'max_features': 'sqrt'},
            0.22713854310135304, 13, 0.20142843988664705, 'sqrt'],
        [np.array([1, 2]),
            {'n_estimators': 14, 'min_samples_leaf': 0.1147302385573586, 'max_features': 'sqrt'},
            0.22734290011048722, 14, 0.1147302385573586, 'sqrt'],
        [np.array([1, 2]),
            {'n_estimators': 17, 'min_samples_leaf': 0.19325882509735576, 'max_features': 'sqrt'},
            0.2279790315504743, 17, 0.19325882509735576, 'sqrt']], dtype=object),
        columns=['lags', 'params', 'mean_absolute_error', 'n_estimators', 'min_samples_leaf', 'max_features'],
        index=pd.Index([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype='int64')
    ).astype({
        'mean_absolute_error': float, 
        'n_estimators': int, 
        'min_samples_leaf': float
    })

    results = results.astype({
        'mean_absolute_error': float, 
        'n_estimators': int, 
        'min_samples_leaf': float
    })

    pd.testing.assert_frame_equal(results, expected_results)


def test_results_output_bayesian_search_optuna_window_features_ForecasterDirect():
    """
    Test output of _bayesian_search_optuna in ForecasterDirect including
    window_features with mocked (mocked done in Skforecast v0.4.3).
    """
    window_features = RollingFeatures(
        stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation'],
        window_sizes = 3,
    )
    forecaster = ForecasterDirect(
                     estimator = RandomForestRegressor(random_state=123),
                     steps     = 3,
                     lags      = 4,
                     window_features = window_features,
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):
        search_space  = {
            'n_estimators'    : trial.suggest_int('n_estimators', 10, 20),
            'min_samples_leaf': trial.suggest_float('min_samples_leaf', 0.1, 1., log=True),
            'max_features'    : trial.suggest_categorical('max_features', ['log2', 'sqrt']),
            'lags'            : trial.suggest_categorical('lags', [2, 4])
        } 
        
        return search_space

    results = _bayesian_search_optuna(
                  forecaster   = forecaster,
                  y            = y,
                  cv           = cv,
                  search_space = search_space,
                  metric       = 'mean_absolute_error',
                  n_trials     = 10,
                  random_state = 123,
                  return_best  = False,
                  verbose      = False
              )[0]
    
    expected_results = pd.DataFrame(
        {
            'lags': [
                np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2]),
                np.array([1, 2]),
                np.array([1, 2]),
            ],
            'params': [
                {
                    'n_estimators': 20,
                    'min_samples_leaf': 0.4839825891759374,
                    'max_features': 'log2',
                },
                {
                    'n_estimators': 15,
                    'min_samples_leaf': 0.41010449151752726,
                    'max_features': 'sqrt',
                },
                {
                    'n_estimators': 14,
                    'min_samples_leaf': 0.782328520465639,
                    'max_features': 'log2',
                },
                {
                    'n_estimators': 17,
                    'min_samples_leaf': 0.21035794225904136,
                    'max_features': 'log2',
                },
                {
                    'n_estimators': 15,
                    'min_samples_leaf': 0.34027307604369605,
                    'max_features': 'sqrt',
                },
                {
                    'n_estimators': 13,
                    'min_samples_leaf': 0.2599119286878713,
                    'max_features': 'log2',
                },
                {
                    'n_estimators': 13,
                    'min_samples_leaf': 0.20142843988664705,
                    'max_features': 'sqrt',
                },
                {
                    'n_estimators': 14,
                    'min_samples_leaf': 0.1147302385573586,
                    'max_features': 'sqrt',
                },
                {
                    'n_estimators': 17,
                    'min_samples_leaf': 0.19325882509735576,
                    'max_features': 'sqrt',
                },
                {
                    'n_estimators': 11,
                    'min_samples_leaf': 0.2714570796881701,
                    'max_features': 'sqrt',
                },
            ],
            'mean_absolute_error': [
                0.2130714159765625,
                0.21370598876262625,
                0.21466582599386722,
                0.21724682621084832,
                0.21740193466522548,
                0.22083884690087485,
                0.22198405983119776,
                0.2246121914126593,
                0.22842351518266302,
                0.22906678546095408,
            ],
            'n_estimators': [20, 15, 14, 17, 15, 13, 13, 14, 17, 11],
            'min_samples_leaf': [
                0.4839825891759374,
                0.41010449151752726,
                0.782328520465639,
                0.21035794225904136,
                0.34027307604369605,
                0.2599119286878713,
                0.20142843988664705,
                0.1147302385573586,
                0.19325882509735576,
                0.2714570796881701,
            ],
            'max_features': [
                'log2',
                'sqrt',
                'log2',
                'log2',
                'sqrt',
                'log2',
                'sqrt',
                'sqrt',
                'sqrt',
                'sqrt',
            ],
        }
    )

    pd.testing.assert_frame_equal(results, expected_results)

    
def test_bayesian_search_optuna_output_file():
    """ 
    Test output file of _bayesian_search_optuna.
    """

    forecaster = ForecasterRecursive(
                     estimator = RandomForestRegressor(random_state=123),
                     lags      = 2 
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    def search_space(trial):
        search_space  = {
            'n_estimators'    : trial.suggest_int('n_estimators', 10, 20),
            'min_samples_leaf': trial.suggest_float('min_samples_leaf', 0.1, 1., log=True),
            'max_features'    : trial.suggest_categorical('max_features', ['log2', 'sqrt']),
            'lags'            : trial.suggest_categorical('lags', [2, 4])
        } 
        
        return search_space

    output_file = 'test_bayesian_search_optuna_output_file.txt'
    _ = _bayesian_search_optuna(
            forecaster   = forecaster,
            y            = y,
            cv           = cv,
            search_space = search_space,
            metric       = 'mean_absolute_error',
            n_trials     = 10,
            random_state = 123,
            return_best  = False,
            verbose      = False,
            output_file  = output_file
        )[0]

    assert os.path.isfile(output_file)
    os.remove(output_file)


@pytest.mark.parametrize("initial_train_size", 
                         [450, '2020-01-19 17:00:00', pd.to_datetime('2020-01-19 17:00:00')], 
                         ids=lambda initial_train_size: f'initial_train_size: {initial_train_size}')
@pytest.mark.parametrize(
        "forecaster",
        [
            ForecasterRecursive(
                estimator=Ridge(random_state=678),
                lags=3,
                transformer_y=None,
                forecaster_id='Recursive_no_transformer'
            ),
            ForecasterDirect(
                estimator=Ridge(random_state=678),
                steps=1,
                lags=3,
                transformer_y=None,
                forecaster_id='Direct_no_transformer'
            ),
            ForecasterRecursive(
                estimator=Ridge(random_state=678),
                lags=3,
                transformer_y=StandardScaler(),
                transformer_exog=StandardScaler(),
                forecaster_id='Recursive_transformers'
            ),
            ForecasterDirect(
                estimator=Ridge(random_state=678),
                steps=1,
                lags=3,
                transformer_y=StandardScaler(),
                transformer_exog=StandardScaler(),
                forecaster_id='Direct_transformer'
            )
        ],
ids=lambda forecaster: f'forecaster: {forecaster.forecaster_id}')
def test_bayesian_search_optuna_outputs_backtesting_one_step_ahead(
    forecaster, initial_train_size
):
    """
    Test that the outputs of _bayesian_search_optuna are equivalent when
    using backtesting and one-step-ahead.
    """
    metrics = [
        "mean_absolute_error",
        "mean_squared_error",
        mean_absolute_percentage_error,
        mean_absolute_scaled_error,
        root_mean_squared_scaled_error,
    ]

    def search_space(trial):
        search_space  = {
            'alpha': trial.suggest_float('alpha', 1e-2, 1.0),
            'lags': trial.suggest_categorical('lags', [2, 4])
        }
        
        return search_space
    
    cv_backtesnting = TimeSeriesFold(
            steps                 = 1,
            initial_train_size    = initial_train_size,
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    cv_one_step_ahead = OneStepAheadFold(
        initial_train_size    = initial_train_size,
        return_all_indexes    = False,
    )
    
    results_backtesting = _bayesian_search_optuna(
        forecaster   = forecaster,
        y            = y_feature_selection,
        exog         = exog_feature_selection,
        cv           = cv_backtesnting,
        search_space = search_space,
        metric       = metrics,
        n_trials     = 10,
        random_state = 123,
        return_best  = False,
        verbose      = False
    )[0]

    warn_msg = re.escape(
        "One-step-ahead predictions are used for faster model comparison, but they "
        "may not fully represent multi-step prediction performance. It is recommended "
        "to backtest the final model for a more accurate multi-step performance "
        "estimate."
    )
    with pytest.warns(OneStepAheadValidationWarning, match = warn_msg):
        results_one_step_ahead = _bayesian_search_optuna(
            forecaster   = forecaster,
            y            = y_feature_selection,
            exog         = exog_feature_selection,
            cv           = cv_one_step_ahead,
            search_space = search_space,
            metric       = metrics,
            n_trials     = 10,
            random_state = 123,
            return_best  = False,
            verbose      = False
        )[0]

    pd.testing.assert_frame_equal(results_backtesting, results_one_step_ahead)



# Unit test _recursive_predict ForecasterRecursive
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import Ridge
from sklearn.linear_model import LinearRegression
from lightgbm import LGBMRegressor
from skforecast.preprocessing import RollingFeatures
from skforecast.recursive import ForecasterRecursive

# Fixtures
from .fixtures_forecaster_recursive import y
from .fixtures_forecaster_recursive import exog
from .fixtures_forecaster_recursive import exog_predict


def test_recursive_predict_output_when_estimator_is_LinearRegression():
    """
    Test _recursive_predict output when using LinearRegression as estimator.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(50)))

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=5)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 5,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array([50., 51., 52., 53., 54.])
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_when_estimator_is_Ridge_StandardScaler():
    """
    Test _recursive_predict output when using Ridge as estimator and
    StandardScaler.
    """
    forecaster = ForecasterRecursive(
                     estimator     = Ridge(random_state=123),
                     lags          = [1, 5],
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=pd.Series(np.arange(50), name='y'))

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=5)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 5,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array([1.745476, 1.803196, 1.865844, 1.930923, 1.997202])

    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_window_features():
    """
    Test _recursive_predict output with window features.
    """
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=4)
    forecaster = ForecasterRecursive(
        LGBMRegressor(verbose=-1), lags=3, window_features=rolling
    )
    forecaster.fit(y=y, exog=exog)

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=10, exog=exog_predict)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 10,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array(
                   [0.584584, 0.487441, 0.483098, 0.483098, 0.580241, 
                    0.584584, 0.584584, 0.487441, 0.483098, 0.483098]
               )
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_two_window_features():
    """
    Test _recursive_predict output with 2 window features.
    """
    rolling = RollingFeatures(stats=['mean'], window_sizes=4)
    rolling_2 = RollingFeatures(stats=['median'], window_sizes=4)
    forecaster = ForecasterRecursive(
        LGBMRegressor(verbose=-1), lags=3, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y, exog=exog)

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=10, exog=exog_predict)
    )
    predictions = forecaster._recursive_predict(
                      steps              = 10,
                      last_window_values = last_window_values,
                      exog_values        = exog_values
                  )
    
    expected = np.array(
                   [0.584584, 0.487441, 0.483098, 0.483098, 0.580241, 
                    0.584584, 0.584584, 0.487441, 0.483098, 0.483098]
               )
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_residuals_zero():
    """
    Test _recursive_predict output with residuals when all residuals are zero.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(50)))

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=5)
    )
    residuals = np.array([[0], [0], [0], [0], [0]])
    predictions = forecaster._recursive_predict(
                      steps                = 5,
                      last_window_values   = last_window_values,
                      exog_values          = exog_values,
                      residuals            = residuals,
                      use_binned_residuals = False
                  )
    
    expected = np.array([50., 51., 52., 53., 54.])

    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_residuals_last_step():
    """
    Test _recursive_predict output with residuals when all residuals are zero 
    except the last step.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(50)))

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=5)
    )
    residuals = np.array([[0], [0], [0], [0], [100]])
    predictions = forecaster._recursive_predict(
                      steps                = 5,
                      last_window_values   = last_window_values,
                      exog_values          = exog_values,
                      residuals            = residuals,
                      use_binned_residuals = False
                  )
    
    expected = np.array([50., 51., 52., 53., 154.])
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_residuals():
    """
    Test _recursive_predict output with residuals.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(50)))

    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=5)
    )
    residuals = np.array([[10], [20], [30], [40], [50]])
    predictions = forecaster._recursive_predict(
                      steps                = 5,
                      last_window_values   = last_window_values,
                      exog_values          = exog_values,
                      residuals            = residuals,
                      use_binned_residuals = False,
                  )
    
    expected = np.array([60., 74.333333, 93.111111, 117.814815, 147.08642])
    
    np.testing.assert_array_almost_equal(predictions, expected)


def test_recursive_predict_output_with_binned_residuals():
    """
    Test _recursive_predict output with binned residuals.
    """
    rng = np.random.default_rng(12345)
    steps = 10
    forecaster = ForecasterRecursive(LGBMRegressor(verbose=-1), lags=3)
    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    last_window_values, exog_values, _, _ = (
        forecaster._create_predict_inputs(steps=steps, exog=exog_predict)
    )

    sampled_residuals = {
        k: v[rng.integers(low=0, high=len(v), size=steps)]
        for k, v in forecaster.in_sample_residuals_by_bin_.items()
    }

    predictions = forecaster._recursive_predict(
                      steps                = steps,
                      last_window_values   = last_window_values,
                      exog_values          = exog_values,
                      residuals            = sampled_residuals,
                      use_binned_residuals = True,
                  )

    expected = np.array(
        [
            0.722443,
            0.849432,
            0.611024,
            0.893993,
            0.612895,
            0.223093,
            0.642686,
            0.68483,
            0.321592,
            0.499459,
        ]
    )

    np.testing.assert_array_almost_equal(predictions, expected)



# Unit test QantityBinner class
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import KBinsDiscretizer
from ...preprocessing import QuantileBinner


def test_QuantileBinner_validate_params():
    """
    Test RollingFeatures _validate_params method.
    """

    params = {
        0: {'n_bins': 1, 'method': 'linear', 'dtype': np.float64, 'subsample': 20000, 'random_state': 123},
        1: {'n_bins': 5, 'method': 'not_valid_method', 'dtype': np.float64, 'subsample': 20000, 'random_state': 123},
        2: {'n_bins': 5, 'method': 'linear', 'dtype': 'not_valid_dtype', 'subsample': 20000, 'random_state': 123},
        3: {'n_bins': 5, 'method': 'linear', 'dtype': np.float64, 'subsample': 'not_int', 'random_state': 123},
        4: {'n_bins': 5, 'method': 'linear', 'dtype': np.float64, 'subsample': 20000, 'random_state': 'not_int'},
    }
        
    err_msg = re.escape(
        f"`n_bins` must be an int greater than 1. Got {params[0]['n_bins']}."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        QuantileBinner(**params[0])

    valid_methods = [
        "inverse_cdf",
        "averaged_inverse_cdf",
        "closest_observation",
        "interpolated_inverse_cdf",
        "hazen",
        "weibull",
        "linear",
        "median_unbiased",
        "normal_unbiased",
    ]
    err_msg = re.escape(
        f"`method` must be one of {valid_methods}. Got {params[1]['method']}."
    )
    with pytest.raises(ValueError, match = err_msg):
        QuantileBinner(**params[1])

    err_msg = re.escape(
        f"`dtype` must be a valid numpy dtype. Got {params[2]['dtype']}."
    )
    with pytest.raises(ValueError, match = err_msg):
        QuantileBinner(**params[2])
    
    err_msg = re.escape(
        f"`subsample` must be an integer greater than or equal to 1. "
        f"Got {params[3]['subsample']}."
    )
    with pytest.raises(ValueError, match = err_msg):
        QuantileBinner(**params[3])
    
    err_msg = re.escape(
        f"`random_state` must be an integer greater than or equal to 0. "
        f"Got {params[4]['random_state']}."
    )
    with pytest.raises(ValueError, match = err_msg):
        QuantileBinner(**params[4])


def test_QuantileBinner_fit_ValueError_when_input_data_is_empty():
    """
    Test ValueError is raised when input data is empty during fit.
    """
    
    X = np.array([])
    binner = QuantileBinner(
        n_bins=10,
        method='linear',
        dtype=np.float64,
        random_state=789654,
    )
    
    err_msg = re.escape("Input data `X` cannot be empty.")
    with pytest.raises(ValueError, match = err_msg):
        binner.fit(X)


def test_QuantileBinner_transform_NotFittedError():
    """
    Test NotFittedError is raised when transform is used before fitting.
    """
    
    X = np.array([])
    binner = QuantileBinner(
        n_bins=10,
        method='linear',
        dtype=np.float64,
        random_state=789654,
    )
    
    err_msg = re.escape(
        "The model has not been fitted yet. Call 'fit' with training data first."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        binner.transform(X)


def test_QuantileBinner_set_params():
    """
    Test set_params method.
    """
        
    binner = QuantileBinner(
        n_bins=10,
        method='linear',
        dtype=np.float64,
        random_state=789654,
    )
    
    params = {
        'n_bins': 5,
        'method': 'inverse_cdf',
        'dtype': np.float32,
        'random_state': 123456,
    }
    
    binner.set_params(**params)
    
    assert binner.n_bins == params['n_bins']
    assert binner.method == params['method']
    assert binner.dtype == params['dtype']
    assert binner.random_state == params['random_state']


def test_QuantileBinner_fit_with_subsample():
    """
    Test QuantileBinner fit method with subsample.
    """
    
    X = np.arange(1000)
    binner = QuantileBinner(
        n_bins=10,
        method='linear',
        subsample=10,
        dtype=np.float64,
        random_state=789654,
    )
    
    binner.fit(X.reshape(-1, 1))

    expected_bin_edges_ = np.array(
        [21., 159.6, 283.8, 388.7, 429.8, 578., 725., 731.6, 734.2,
         739.4, 743.]
    )
    expected_n_bins_ = 10
    expected_intervals_ = {
        0: (21.0, 159.6),
        1: (159.6, 283.8),
        2: (283.8, 388.7),
        3: (388.7, 429.8),
        4: (429.8, 578.0),
        5: (578.0, 725.0),
        6: (725.0, 731.6),
        7: (731.6, 734.2),
        8: (734.2, 739.4),
        9: (739.4, 743.0)
    }

    np.testing.assert_array_almost_equal(binner.bin_edges_, expected_bin_edges_)
    assert binner.n_bins_ == expected_n_bins_
    assert binner.intervals_ == expected_intervals_


def test_QuantileBinner_is_equivalent_to_KBinsDiscretizer():
    """
    Test that QuantileBinner is equivalent to KBinsDiscretizer when `method='linear'`.
    """
    
    X = np.random.normal(10, 10, 10000)
    n_bins_grid = [2, 10, 20]

    for n_bins in n_bins_grid:
        binner_1 = KBinsDiscretizer(
            n_bins=n_bins,
            encode="ordinal",
            strategy="quantile",
            dtype=np.float64,
            random_state=789654,
        )
        binner_2 = QuantileBinner(
            n_bins=n_bins,
            method='linear',
            dtype=np.float64,
            random_state=789654,
        )

        binner_1.fit(X.reshape(-1, 1))
        transformed_1 = binner_1.transform(X.reshape(-1, 1)).flatten()
        transformed_2 = binner_2.fit_transform(X)

        np.testing.assert_array_almost_equal(binner_1.bin_edges_[0], binner_2.bin_edges_)
        np.testing.assert_array_almost_equal(transformed_1, transformed_2)
